import unittest
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import responses
import yaml
from sqlalchemy import create_engine

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.salesforce_api.org_schema import Base, Schema
from cumulusci.salesforce_api.org_schema_models import Field as SQLField
from cumulusci.salesforce_api.org_schema_models import SObject as SQLSObject
from cumulusci.tasks.bulkdata import GenerateMapping
from cumulusci.tasks.bulkdata.tests.utils import _make_task
from cumulusci.utils import temporary_dir


@contextmanager
def _temp_schema_for_tests(describe_data: list):
    with TemporaryDirectory() as tempdir:
        tempfile = str(Path(tempdir) / "tempdb.db")
        engine = create_engine(f"sqlite:///{tempfile}")
        Base.metadata.bind = engine
        Base.metadata.create_all()
        schema = Schema(engine, f"{tempfile}.gz")
        try:
            date = "Thu, 09 Feb 2021 21:35:07 GMT"
            describe_data_with_dates = [(dct, date) for dct in describe_data]
            schema._populate_cache_from_describe(describe_data_with_dates, date)
            yield schema
        finally:
            schema.close()


def _SObject(name, fields, **kwargs):
    assert name
    if isinstance(fields, list):
        fields = {field["name"]: field for field in fields}
    return SQLSObject(name=name, fields=fields, **kwargs)


def _Field(name: str, dct=None, **kwargs):
    dct = dct or {}
    kwargs.update(kwargs)
    return SQLField(name=name, **kwargs)


class TestMappingGenerator(unittest.TestCase):
    def test_defaults_options(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertEqual([], t.options["ignore"])
        self.assertEqual("", t.options["namespace_prefix"])
        self.assertEqual("ask", t.options["break_cycles"])
        self.assertEqual([], t.options["include"])

    def test_postfixes_underscores_to_namespace(self):
        t = _make_task(
            GenerateMapping, {"options": {"namespace_prefix": "t", "path": "t"}}
        )

        self.assertEqual("t__", t.options["namespace_prefix"])

    def test_splits_ignore_string(self):
        t = _make_task(
            GenerateMapping, {"options": {"ignore": "Account, Contact", "path": "t"}}
        )

        self.assertEqual(["Account", "Contact"], t.options["ignore"])

    def test_accepts_ignore_list(self):
        t = _make_task(
            GenerateMapping,
            {"options": {"ignore": ["Account", "Contact"], "path": "t"}},
        )

        self.assertEqual(["Account", "Contact"], t.options["ignore"])

    def test_accepts_include_list(self):
        t = _make_task(
            GenerateMapping, {"options": {"include": ["Foo", "Bar"], "path": "t"}}
        )

        self.assertEqual(["Foo", "Bar"], t.options["include"])

    @responses.activate
    def test_checks_include_list(self):
        t = _make_task(
            GenerateMapping, {"options": {"include": ["Foo", "Bar"], "path": "t"}}
        )
        t.project_config.project__package__api_version = "45.0"

        with self._prepare_describe_mock(t, {}):
            t._init_task()

        with pytest.raises(TaskOptionsError):
            t._collect_objects({})

    def test_is_any_custom_api_name(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_any_custom_api_name("Custom__c"))
        self.assertFalse(t._is_any_custom_api_name("Standard"))

    def test_is_our_custom_api_name(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_our_custom_api_name("Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("Standard"))
        self.assertFalse(t._is_our_custom_api_name("t__Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("f__Custom__c"))

        t.options["namespace_prefix"] = "t__"
        self.assertTrue(t._is_our_custom_api_name("Custom__c"))
        self.assertTrue(t._is_our_custom_api_name("t__Custom__c"))
        self.assertFalse(t._is_our_custom_api_name("f__Custom__c"))

    def test_is_core_field(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        self.assertTrue(t._is_core_field("Name"))
        self.assertFalse(t._is_core_field("Custom__c"))

    def test_is_object_mappable(self):
        t = _make_task(GenerateMapping, {"options": {"ignore": "Account", "path": "t"}})

        self.assertTrue(
            t._is_object_mappable({"name": "Contact", "customSetting": False})
        )
        self.assertFalse(
            t._is_object_mappable({"name": "Account", "customSetting": False})
        )
        self.assertFalse(
            t._is_object_mappable(
                {"name": "Contact__ChangeEvent", "customSetting": False}
            )
        )
        self.assertFalse(
            t._is_object_mappable({"name": "Custom__c", "customSetting": True})
        )

    def test_is_field_mappable(self):
        t = _make_task(
            GenerateMapping, {"options": {"ignore": "Account.ParentId", "path": "t"}}
        )

        t.mapping_objects = ["Account", "Contact"]

        self.assertTrue(
            t._is_field_mappable(
                "Account",
                {"name": "Name", "type": "string", "label": "Name", "createable": True},
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {"name": "Name", "type": "base64", "label": "Name", "createable": True},
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "Name",
                    "type": "string",
                    "label": "Name (Deprecated)",
                    "createable": True,
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "ParentId",
                    "type": "reference",
                    "label": "Parent",
                    "createable": True,
                    "referenceTo": ["Account"],
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Account",
                {
                    "name": "Name",
                    "type": "string",
                    "label": "Name",
                    "createable": False,
                },
            )
        )
        self.assertFalse(
            t._is_field_mappable(
                "Contact",
                {
                    "name": "OwnerId",
                    "type": "reference",
                    "label": "Owner",
                    "createable": True,
                    "referenceTo": ["User", "Group"],
                },
            )
        )

    def test_has_our_custom_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        sobject = _SObject("Blah", [_Field(name="CustomObject__c")])
        self.assertTrue(t._has_our_custom_fields(sobject))
        self.assertTrue(
            t._has_our_custom_fields(
                _SObject("Blah", [_Field("Custom__c"), _Field("Standard")])
            )
        )
        self.assertFalse(
            t._has_our_custom_fields(_SObject("Blah", [_Field("Standard")]))
        )
        self.assertFalse(t._has_our_custom_fields(_SObject("Blah", [])))

    def test_is_lookup_to_included_object(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account"]

        self.assertTrue(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Account"]}
            )
        )
        self.assertFalse(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Contact"]}
            )
        )
        self.assertFalse(
            t._is_lookup_to_included_object(
                {"type": "reference", "referenceTo": ["Account", "Contact"]}
            )
        )

    @contextmanager
    def _prepare_describe_mock(self, task, describe_data):
        with mock.patch(
            "cumulusci.tasks.bulkdata.generate_mapping.get_org_schema",
            lambda *args, **kwargs: _temp_schema_for_tests(describe_data),
        ):
            yield

    def _mock_field(self, name, field_type="string", **kwargs):
        field_data = {
            "name": name,
            "type": field_type,
            "createable": True,
            "nillable": True,
            "label": name,
            **kwargs,
        }
        return field_data

    @responses.activate
    def test_run_task(self):
        t = _make_task(GenerateMapping, {"options": {"path": "mapping.yaml"}})
        t.project_config.project__package__api_version = "45.0"
        describe_data = [
            {
                "name": "Parent",
                "fields": [self._mock_field("Id"), self._mock_field("Custom__c")],
            },
            {
                "name": "Child__c",
                "fields": [
                    self._mock_field("Id"),
                    self._mock_field(
                        "Account__c",
                        field_type="reference",
                        referenceTo=["Parent"],
                        relationshipOrder=None,
                    ),
                ],
            },
        ]

        with temporary_dir(), self._prepare_describe_mock(t, describe_data):
            t()

            with open("mapping.yaml", "r") as fh:
                content = yaml.safe_load(fh)

            self.assertEqual(["Insert Parent", "Insert Child__c"], list(content.keys()))
            self.assertEqual("Parent", t.mapping["Insert Parent"]["sf_object"])
            self.assertEqual(["Custom__c"], t.mapping["Insert Parent"]["fields"])

            self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
            assert "fields" not in t.mapping["Insert Child__c"]
            self.assertEqual(
                ["Account__c"], list(t.mapping["Insert Child__c"]["lookups"].keys())
            )
            self.assertEqual(
                "Parent", t.mapping["Insert Child__c"]["lookups"]["Account__c"]["table"]
            )

    @responses.activate
    def test_collect_objects__simple_custom_objects(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Contact", "fields": [self._mock_field("Name")]},
            {
                "name": "Custom__c",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "User", "fields": [self._mock_field("Name")]},
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)

        self.assertEqual(set(["Account", "Custom__c"]), set(t.mapping_objects))

    @responses.activate
    def test_collect_objects__force_include_objects(self):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "include": ["Contact", "User"]}}
        )
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Contact", "fields": [self._mock_field("Name")]},
            {
                "name": "Custom__c",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "User", "fields": [self._mock_field("Name")]},
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)

        self.assertEqual(
            set(["Account", "Custom__c", "Contact", "User"]), set(t.mapping_objects)
        )

    @responses.activate
    def test_collect_objects__force_include_objects__already_included(self):
        t = _make_task(
            GenerateMapping,
            {"options": {"path": "t", "include": ["Contact", "Custom__c"]}},
        )
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Contact", "fields": [self._mock_field("Name")]},
            {
                "name": "Custom__c",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "User", "fields": [self._mock_field("Name")]},
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)
        assert len(t.mapping_objects) == 3

        self.assertEqual(
            set(["Account", "Custom__c", "Contact"]), set(t.mapping_objects)
        )

    @responses.activate
    def test_collect_objects__custom_lookup_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Contact", "fields": [self._mock_field("Name")]},
            {
                "name": "Custom__c",
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "Lookup__c",
                        field_type="reference",
                        relationshipOrder=None,
                        referenceTo=["Contact"],
                    ),
                ],
            },
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)

        self.assertEqual(
            set(["Account", "Custom__c", "Contact"]), set(t.mapping_objects)
        )

    @responses.activate
    def test_collect_objects__master_detail_fields(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Opportunity", "fields": [self._mock_field("Name")]},
            {
                "name": "OpportunityLineItem",
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "OpportunityId",
                        field_type="reference",
                        relationshipOrder=1,
                        referenceTo=["Opportunity"],
                    ),
                ],
            },
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)

        self.assertEqual(
            set(["Account", "OpportunityLineItem", "Opportunity"]),
            set(t.mapping_objects),
        )

    @responses.activate
    def test_collect_objects__duplicate_references(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__api_version = "45.0"

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Custom__c")],
            },
            {"name": "Opportunity", "fields": [self._mock_field("Name")]},
            {
                "name": "OpportunityLineItem",
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Custom__c"),
                    self._mock_field(
                        "OpportunityId",
                        field_type="reference",
                        relationshipOrder=1,
                        referenceTo=["Opportunity"],
                    ),
                    self._mock_field(
                        "CustomLookup__c",
                        field_type="reference",
                        relationshipOrder=None,
                        referenceTo=["Opportunity"],
                    ),
                ],
            },
        ]

        with _temp_schema_for_tests(describe_data) as schema:
            t._init_task()
            t._collect_objects(schema)

        self.assertEqual(
            set(["Account", "OpportunityLineItem", "Opportunity"]),
            set(t.mapping_objects),
        )

    def test_simplify_schema(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity", "Child__c"]
        stage_name = self._mock_field("StageName", nillable=False)

        describe_data = [
            {
                "name": "Account",
                "fields": [self._mock_field("Name"), self._mock_field("Industry")],
            },
            {"name": "Opportunity", "fields": [self._mock_field("Name"), stage_name]},
            {
                "name": "Child__c",
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field("Test__c"),
                    self._mock_field("Attachment__c", field_type="base64"),
                ],
            },
        ]
        with _temp_schema_for_tests(describe_data) as org_schema:
            t._simplify_schema(org_schema)

        assert set(t.simple_schema.keys()) == {"Account", "Opportunity", "Child__c"}
        assert set(t.simple_schema["Account"].keys()) == {"Name"}
        assert t.simple_schema["Account"]["Name"].name == "Name"
        assert set(t.simple_schema["Opportunity"].keys()) == {"Name", "StageName"}
        assert t.simple_schema["Opportunity"]["Name"].name == "Name"
        assert t.simple_schema["Opportunity"]["StageName"].name == "StageName"
        assert t.simple_schema["Opportunity"]["StageName"].nillable is False
        assert set(t.simple_schema["Child__c"].keys()) == {"Name", "Test__c"}
        assert t.simple_schema["Child__c"]["Name"].name == "Name"
        assert t.simple_schema["Child__c"]["Test__c"].name == "Test__c"

    def test_simplify_schema__tracks_references(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity"]
        account_id = self._mock_field(
            "AccountId",
            field_type="reference",
            referenceTo=["Account"],
            relationshipOrder=1,
        )
        describe_data = [
            {"name": "Account", "fields": [self._mock_field("Name")]},
            {"name": "Opportunity", "fields": [self._mock_field("Name"), account_id]},
        ]
        with _temp_schema_for_tests(describe_data) as org_schema:
            t._simplify_schema(org_schema)

        assert list(t.refs.keys()) == ["Opportunity"]
        referents = list(t.refs.values())[0]
        assert list(referents.keys()) == ["Account"]
        account_sobj = list(referents.values())[0]
        assert list(account_sobj.keys()) == ["AccountId"]
        account_id_field = list(account_sobj.values())[0]
        assert account_id_field.name == "AccountId"

    def test_simplify_schema__includes_recordtypeid(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Opportunity"]
        describe_data = [
            {"name": "Account", "fields": [self._mock_field("Name")]},
            {
                "name": "Opportunity",
                "fields": [
                    self._mock_field("Name"),
                    self._mock_field(
                        "AccountId",
                        field_type="reference",
                        referenceTo=["Account"],
                        relationshipOrder=1,
                    ),
                    self._mock_field("RecordTypeId"),
                ],
                "recordTypeInfos": [{"Name": "Master"}, {"Name": "Donation"}],
            },
        ]
        with _temp_schema_for_tests(describe_data) as org_schema:
            t._simplify_schema(org_schema)
        self.assertIn("RecordTypeId", t.simple_schema["Opportunity"])
        self.assertNotIn("RecordTypeId", t.simple_schema["Account"])

    @mock.patch("click.prompt")
    def test_build_mapping(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        prompt.return_value = "Account"

        t.simple_schema = {
            "Account": {
                "Name": self._mock_field("Name"),
                "Dependent__c": self._mock_field(
                    "Dependent__c", field_type="reference", referenceTo=["Child__c"]
                ),
            },
            "Child__c": {
                "Name": self._mock_field("Name"),
                "Account__c": self._mock_field(
                    "Account__c", field_type="reference", referenceTo=["Account"]
                ),
                "Self__c": self._mock_field(
                    "Self__c", field_type="reference", referenceTo=["Child__c"]
                ),
            },
        }
        t.refs = {
            "Child__c": {
                "Account": {"Account__c": _Field("Account__c", {"nillable": True})}
            },
            "Account": {
                "Child__c": {"Dependent__c": _Field("Dependent__c", {"nillable": True})}
            },
        }

        t._build_mapping()
        self.assertEqual(["Insert Account", "Insert Child__c"], list(t.mapping.keys()))
        self.assertEqual("Account", t.mapping["Insert Account"]["sf_object"])
        self.assertEqual(["Name"], t.mapping["Insert Account"]["fields"])
        self.assertEqual(
            ["Dependent__c"], list(t.mapping["Insert Account"]["lookups"].keys())
        )
        self.assertEqual(
            "Child__c", t.mapping["Insert Account"]["lookups"]["Dependent__c"]["table"]
        )

        self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
        self.assertEqual(["Name"], t.mapping["Insert Child__c"]["fields"])
        self.assertEqual(
            ["Account__c", "Self__c"],
            list(t.mapping["Insert Child__c"]["lookups"].keys()),
        )
        self.assertEqual(
            "Account", t.mapping["Insert Child__c"]["lookups"]["Account__c"]["table"]
        )
        self.assertEqual(
            "Child__c", t.mapping["Insert Child__c"]["lookups"]["Self__c"]["table"]
        )

    @mock.patch("click.prompt")
    def test_build_mapping__strip_namespace(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__namespace = "ns"
        prompt.return_value = "ns__Parent__c"

        t.simple_schema = {
            "ns__Parent__c": {
                "Name": self._mock_field("Name"),
                "ns__Dependent__c": self._mock_field(
                    "ns__Dependent__c",
                    field_type="reference",
                    referenceTo=["ns__Child__c"],
                ),
            },
            "ns__Child__c": {
                "Name": self._mock_field("Name"),
                "ns__Parent__c": self._mock_field(
                    "ns__Parent__c",
                    field_type="reference",
                    referenceTo=["ns__Parent__c"],
                ),
                "ns__Self__c": self._mock_field(
                    "ns__Self__c", field_type="reference", referenceTo=["ns__Child__c"]
                ),
            },
        }
        t.refs = {
            "ns__Child__c": {
                "ns__Parent__c": {
                    "ns__Parent__c": _Field("ns__Parent__c", {"nillable": False})
                }
            },
            "ns__Parent__c": {
                "ns__Child__c": {
                    "ns__Dependent__c": _Field("ns__Dependent__c", {"nillable": True})
                }
            },
        }

        t._build_mapping()
        self.assertEqual(
            ["Insert Parent__c", "Insert Child__c"], list(t.mapping.keys())
        )
        self.assertEqual("Parent__c", t.mapping["Insert Parent__c"]["sf_object"])
        self.assertEqual(["Name"], t.mapping["Insert Parent__c"]["fields"])
        self.assertEqual(
            ["Dependent__c"], list(t.mapping["Insert Parent__c"]["lookups"].keys())
        )
        self.assertEqual(
            "Child__c",
            t.mapping["Insert Parent__c"]["lookups"]["Dependent__c"]["table"],
        )

        self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
        self.assertEqual(["Name"], t.mapping["Insert Child__c"]["fields"])
        self.assertEqual(
            ["Parent__c", "Self__c"],
            list(t.mapping["Insert Child__c"]["lookups"].keys()),
        )
        self.assertEqual(
            "Parent__c", t.mapping["Insert Child__c"]["lookups"]["Parent__c"]["table"]
        )
        self.assertEqual(
            "Child__c", t.mapping["Insert Child__c"]["lookups"]["Self__c"]["table"]
        )

    @mock.patch("click.prompt")
    def test_build_mapping__no_strip_namespace_if_dup_component(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})
        t.project_config.project__package__namespace = "ns"
        prompt.return_value = "ns__Parent__c"

        t.simple_schema = {
            "ns__Parent__c": {"Name": self._mock_field("Name")},
            "ns__Child__c": {
                "Name": self._mock_field("Name"),
                "Test__c": self._mock_field("Test__c"),
                "ns__Test__c": self._mock_field("ns__Test__c"),
                "ns__Parent__c": self._mock_field(
                    "ns__Parent__c",
                    field_type="reference",
                    referenceTo=["ns__Parent__c"],
                ),
                "Parent__c": self._mock_field(
                    "Parent__c", field_type="reference", referenceTo=["ns__Child__c"]
                ),
            },
            "Child__c": {"Name": self._mock_field("Name")},
        }
        t.refs = {
            "ns__Child__c": {
                "ns__Parent__c": {"ns__Parent__c": _Field("ns__Parent__c", {})}
            }
        }

        t._build_mapping()

        self.assertEqual(
            set(["Insert Parent__c", "Insert ns__Child__c", "Insert Child__c"]),
            set(t.mapping.keys()),
        )

        self.assertEqual("ns__Child__c", t.mapping["Insert ns__Child__c"]["sf_object"])
        self.assertEqual(
            ["Name", "Test__c", "ns__Test__c"],
            t.mapping["Insert ns__Child__c"]["fields"],
        )
        self.assertEqual(
            set(["ns__Parent__c", "Parent__c"]),
            set(t.mapping["Insert ns__Child__c"]["lookups"].keys()),
        )
        self.assertEqual(
            "Parent__c",
            t.mapping["Insert ns__Child__c"]["lookups"]["ns__Parent__c"]["table"],
        )
        self.assertEqual(
            "ns__Child__c",
            t.mapping["Insert ns__Child__c"]["lookups"]["Parent__c"]["table"],
        )

        self.assertEqual("Child__c", t.mapping["Insert Child__c"]["sf_object"])
        self.assertEqual(["Name"], t.mapping["Insert Child__c"]["fields"])

    def test_build_mapping__warns_polymorphic_lookups(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        t.mapping_objects = ["Account", "Contact", "Custom__c"]
        t.simple_schema = {
            "Account": {"Name": self._mock_field("Name")},
            "Contact": {"Name": self._mock_field("Name")},
            "Custom__c": {
                "Name": self._mock_field("Name"),
                "PolyLookup__c": self._mock_field(
                    "PolyLookup__c",
                    field_type="reference",
                    referenceTo=["Account", "Contact"],
                ),
            },
        }
        t.refs = {
            "Custom__c": {
                "Account": {"PolyLookup__c": _Field("PolyLookup__c", {})},
                "Contact": {"PolyLookup__c": _Field("PolyLookup__c", {})},
            }
        }
        t.logger = mock.Mock()

        t._build_mapping()
        t.logger.warning.assert_called_once_with(
            "Field Custom__c.PolyLookup__c is a polymorphic lookup, which is not supported"
        )

    def test_split_dependencies__no_cycles(self):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        stack = t._split_dependencies(
            ["Account", "Contact", "Opportunity", "Custom__c"],
            {
                "Contact": {"Account": {"AccountId": _Field("AccountId", {})}},
                "Opportunity": {
                    "Account": {"AccountId": _Field("AccountId", {})},
                    "Contact": {"Primary_Contact__c": _Field("Primary_Contact__c", {})},
                },
                "Custom__c": {
                    "Account": {"Account__c": _Field("Account__c", {})},
                    "Contact": {"Contact__c": _Field("Contact__c", {})},
                    "Opportunity": {"Opp__c": _Field("Opp__c", {})},
                },
            },
        )

        self.assertEqual(["Account", "Contact", "Opportunity", "Custom__c"], stack)

    @mock.patch("click.prompt")
    def test_split_dependencies__interviews_for_cycles(self, prompt):
        t = _make_task(GenerateMapping, {"options": {"path": "t"}})

        prompt.return_value = "Account"

        self.assertEqual(
            ["Custom__c", "Account", "Contact", "Opportunity"],
            t._split_dependencies(
                ["Account", "Contact", "Opportunity", "Custom__c"],
                {
                    "Account": {
                        "Contact": {
                            "Primary_Contact__c": _Field(
                                "Primary_Contact__c", {"nillable": True}
                            )
                        }
                    },
                    "Contact": {
                        "Account": {
                            "AccountId": _Field("AccountId", {"nillable": True})
                        }
                    },
                    "Opportunity": {
                        "Account": {
                            "AccountId": _Field("AccountId", {"nillable": True})
                        },
                        "Contact": {
                            "Primary_Contact__c": _Field(
                                "Primary_Contact__c", {"nillable": True}
                            )
                        },
                    },
                },
            ),
        )

    @mock.patch("click.prompt")
    @mock.patch("random.choice")
    def test_split_dependencies__auto_pick_cycles_priortize_Account(
        self, choice, prompt
    ):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "break_cycles": "auto"}}
        )

        prompt.side_effect = AssertionError("Shouldn't be called")
        choice.side_effect = AssertionError("Shouldn't be called")
        split_dependencies = t._split_dependencies(
            ["Account", "Contact", "Opportunity", "Custom__c"],
            {
                "Account": {
                    "Contact": {
                        "Primary_Contact__c": _Field(
                            "Primary_Contact__c", {"nillable": False}
                        )
                    }
                },
                "Contact": {
                    "Account": {"AccountId": _Field("AccountId", {"nillable": False})}
                },
                "Opportunity": {
                    "Account": {"AccountId": _Field("AccountId", {"nillable": False})},
                    "Contact": {
                        "Primary_Contact__c": _Field(
                            "Primary_Contact__c", {"nillable": False}
                        )
                    },
                },
            },
        )

        self.assertEqual(
            ["Custom__c", "Account", "Contact", "Opportunity"], split_dependencies
        )
        assert not choice.mock_calls

    @mock.patch("click.prompt")
    def test_split_dependencies__auto_pick_cycles_randomly(self, prompt):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "break_cycles": "auto"}}
        )

        prompt.side_effect = AssertionError("Shouldn't be called")
        split_dependencies = t._split_dependencies(
            ["Account", "Contact", "Opportunity", "Custom__c"],
            {
                "Account": {
                    "Custom__c": {
                        "Non_Nillable_Custom__c": _Field(
                            "Non_Nillable_Custom__c", {"nillable": False}
                        )
                    }
                },
                "Custom__c": {
                    "Account": {"AccountId": _Field("AccountId", {"nillable": False})}
                },
            },
        )

        self.assertEqual(
            ["Contact", "Opportunity", "Account", "Custom__c"], split_dependencies
        )

    @mock.patch("click.prompt")
    @mock.patch("random.choice")
    def test_split_dependencies__auto_pick_cycles_by_relationship_type(
        self, random_choice, prompt
    ):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "break_cycles": "auto"}}
        )

        prompt.side_effect = AssertionError("Shouldn't be called")
        random_choice.side_effect = AssertionError("Shouldn't be called")

        split_dependencies = t._split_dependencies(
            ["AccountLike__c", "ContactLike__c", "OpportunityLike__c", "Custom__c"],
            {
                # Primary_Contact__c is not nillable, so ContactLike__c must be loaded before
                # AccountLike__c despite the cycle
                "AccountLike__c": {
                    "ContactLike__c": {
                        "Primary_Contact__c": _Field(
                            "Primary_Contact__c", nillable=False
                        )
                    }
                },
                "ContactLike__c": {
                    "AccountLike__c": {
                        "AccountId": _Field("AccountLike__c", nillable=True)
                    }
                },
                "OpportunityLike__c": {
                    "AccountLike__c": {
                        "AccountId": _Field("AccountLike__c", nillable=True)
                    },
                    "ContactLike__c": {
                        "Primary_Contact__c": _Field("AccountLike__c", nillable=True)
                    },
                },
            },
        )

        self.assertEqual(
            ["Custom__c", "ContactLike__c", "AccountLike__c", "OpportunityLike__c"],
            split_dependencies,
        )
        random_choice.assert_not_called()

    @mock.patch("click.prompt")
    def test_split_dependencies__auto_pick_cycles(self, prompt):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "break_cycles": "auto"}}
        )

        prompt.return_value = AssertionError("Shouldn't be called")

        self.assertEqual(
            set(["Custom__c", "Account", "Contact", "Opportunity"]),
            set(
                t._split_dependencies(
                    ["Account", "Contact", "Opportunity", "Custom__c"],
                    {
                        "Account": {
                            "Contact": {
                                "Primary_Contact__c": _Field(
                                    "Primary_Contact__c", {"nillable": True}
                                )
                            }
                        },
                        "Contact": {
                            "Account": {
                                "AccountId": _Field("AccountId", {"nillable": True})
                            }
                        },
                        "Opportunity": {
                            "Account": {
                                "AccountId": _Field("AccountId", {"nillable": True})
                            },
                            "Contact": {
                                "Primary_Contact__c": _Field(
                                    "Primary_Contact__c", {"nillable": True}
                                )
                            },
                        },
                    },
                )
            ),
        )

    @mock.patch("click.prompt")
    def test_split_dependencies__ask_pick_cycles(self, prompt):
        t = _make_task(
            GenerateMapping, {"options": {"path": "t", "break_cycles": "ask"}}
        )
        prompt.return_value = "Custom__c"

        self.assertEqual(
            set(["Custom__c", "Account", "Contact", "Opportunity"]),
            set(
                t._split_dependencies(
                    ["Account", "Contact", "Opportunity", "Custom__c"],
                    {
                        "Account": {
                            "Custom__c": {
                                "Custom__c": _Field("Custom__c", {"nillable": False})
                            }
                        },
                        "Custom__c": {
                            "Account": {
                                "Account__c": _Field("Account__c", {"nillable": False})
                            }
                        },
                    },
                )
            ),
        )

        prompt.assert_called_once()
        assert prompt.mock_calls

    def test_options_error(self):
        with pytest.raises(TaskOptionsError):
            _make_task(
                GenerateMapping, {"options": {"path": "t", "break_cycles": "foo"}}
            )


@pytest.mark.needs_org()  # too hard to make these VCR-compatible due to data volume
@pytest.mark.slow()
class TestIntegrationGenerateMapping:
    def test_simple_generate(self, create_task):
        "Generate a mapping against a provided org."
        with TemporaryDirectory() as t:
            tempfile = Path(t) / "tempfile.mapping.yml"

            task = create_task(GenerateMapping, {"path": tempfile})
            assert not Path(tempfile).exists()
            task()
            assert Path(tempfile).exists()

    def test_generate_with_cycles(self, create_task):
        "Generate a mapping that necessarily includes some reference cycles"
        with TemporaryDirectory() as t:
            tempfile = Path(t) / "tempfile.mapping.yml"

            task = create_task(
                GenerateMapping,
                {
                    "path": tempfile,
                    "include": [
                        "Account",
                        "Contact",
                        "Opportunity",
                        "OpportunityContactRole",
                    ],
                },
            )
            assert not Path(tempfile).exists()
            task()
            assert Path(tempfile).exists()

    #  May be useful for experimenting with future versions
    # @pytest.mark.vcr()
    # def test_big_generate(self, create_task, sf):
    #     "Generate a large mapping that includes every reachable object"
    #     with TemporaryDirectory() as t:
    #         tempfile = Path(t) / "tempfile.mapping.yml"

    #         # every_obj = [obj["name"] for obj in sf.describe()["sobjects"]]
    #         every_obj = ["Account", ]

    #         task = create_task(
    #             GenerateMapping, {"path": tempfile, "include": every_obj}
    #         )
    #         assert not Path(tempfile).exists()
    #         task()
    #         assert Path(tempfile).exists()



import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from snowfakery import generate_data

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_list_arg, process_list_of_pairs_dict_arg
from cumulusci.tasks.bulkdata.base_generate_data_task import BaseGenerateDataTask


class GenerateDataFromYaml(BaseGenerateDataTask):
    """Generate sample data from a YAML template file."""

    task_docs = """
    Generate YAML from a Snowfakery recipe.
    """

    task_options = {
        **BaseGenerateDataTask.task_options,
        "generator_yaml": {
            "description": "A Snowfakery recipe file to use",
            "required": True,
        },
        "vars": {
            "description": "Pass values to override options in the format VAR1:foo,VAR2:bar"
        },
        "generate_mapping_file": {
            "description": "A path to put a mapping file inferred from the generator_yaml",
            "required": False,
        },
        "num_records_tablename": {
            "description": "A string representing which table determines when the recipe execution is done.",
            "required": False,
        },
        "continuation_file": {
            "description": "YAML file generated by Snowfakery representing next steps for data generation"
        },
        "generate_continuation_file": {
            "description": "Path for Snowfakery to put its next continuation file"
        },
        "working_directory": {
            "description": "Default path for temporary / working files"
        },
        "loading_rules": {
            "description": "Path to .load.yml file containing rules to use to load the file. Defaults to <recipename>.load.yml . Multiple files can be comma separated."
        },
        "num_records": {
            "description": (
                "Target number of records. "
                "You will get at least this many records, but may get more. "
                "The recipe will always execute to completion, so if it creates "
                "3 records per execution and you ask for 5, you will get 6."
            ),
            "required": False,
        },
    }
    stopping_criteria = None

    def __init__(self, *args, **kwargs):
        self.vars = {}
        super().__init__(*args, **kwargs)

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.yaml_file = os.path.abspath(self.options["generator_yaml"])
        if not os.path.exists(self.yaml_file):
            raise TaskOptionsError(f"Cannot find {self.yaml_file}")
        if "vars" in self.options:
            self.vars = process_list_of_pairs_dict_arg(self.options["vars"])
        self.generate_mapping_file = self.options.get("generate_mapping_file")
        if self.generate_mapping_file:
            self.generate_mapping_file = os.path.abspath(self.generate_mapping_file)
        num_records = self.options.get("num_records")
        if num_records is not None:
            num_records = int(num_records)
            num_records_tablename = self.options.get("num_records_tablename")

            if not num_records_tablename:
                raise TaskOptionsError(
                    "Cannot specify num_records without num_records_tablename."
                )

            self.stopping_criteria = (num_records, num_records_tablename)
        self.working_directory = self.options.get("working_directory")
        loading_rules = process_list_arg(self.options.get("loading_rules")) or []
        self.loading_rules = [Path(path) for path in loading_rules if path]

    def _generate_data(self, db_url, mapping_file_path, num_records, current_batch_num):
        """Generate all of the data"""
        if num_records is not None:  # num_records is None means execute Snowfakery once
            self.logger.info(f"Generating batch {current_batch_num} with {num_records}")
        self.generate_data(db_url, num_records, current_batch_num)
        self.logger.info("Generated batch")

    def default_continuation_file_path(self):
        return Path(self.working_directory) / "continuation.yml"

    def get_old_continuation_file(self) -> Optional[Path]:
        """Use a continuation file if specified or look for one in the working directory

        Return None if no file can be found.

        If this code is used within a GenerateAndLoad loop, the continuation files will go
        into the working directory specified by the GenerateAndLoad caller.
        """
        old_continuation_file = self.options.get("continuation_file")

        if old_continuation_file:
            old_continuation_file = Path(old_continuation_file)
            if not old_continuation_file.exists():
                raise TaskOptionsError(f"{old_continuation_file} does not exist")
        elif self.working_directory:
            path = self.default_continuation_file_path()
            if path.exists():
                old_continuation_file = path

        return old_continuation_file

    @contextmanager
    def open_new_continuation_file(self):
        """Create a continuation file based on config or working directory

        Return None if there is no config nor working directory.

        If this code is used within a GenerateAndLoad loop, the continuation files will go
        into the working directory specified by the GenerateAndLoad caller.
        """
        if self.options.get("generate_continuation_file"):
            continuation_path = self.options["generate_continuation_file"]

        elif self.working_directory:
            continuation_path = Path(self.working_directory) / "continuation_next.yml"
        else:
            continuation_path = None

        if continuation_path:
            with open(continuation_path, "w+") as new_continuation_file:
                yield new_continuation_file
        else:
            yield None

    def generate_data(self, dburl, num_records, current_batch_num):
        old_continuation_file = self.get_old_continuation_file()
        if old_continuation_file:
            # reopen to ensure file pointer is at starting point
            old_continuation_file = open(old_continuation_file, "r")
        with self.open_new_continuation_file() as new_continuation_file:
            generate_data(
                yaml_file=self.yaml_file,
                user_options=self.vars,
                target_number=self.stopping_criteria,
                continuation_file=old_continuation_file,
                generate_continuation_file=new_continuation_file,
                generate_cci_mapping_file=self.generate_mapping_file,
                dburl=dburl,
                load_declarations=self.loading_rules,
                should_create_cci_record_type_tables=True,
                plugin_options={
                    "org_config": self.org_config,
                    "project_config": self.project_config,
                },
            )

        if (
            new_continuation_file
            and Path(new_continuation_file.name).exists()
            and self.working_directory
        ):
            shutil.move(
                new_continuation_file.name, self.default_continuation_file_path()
            )



CREATE_SUBSCRIBER_ATTRIBUTE_EXPECTED_SOAP_RESPONSE = """<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
    xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing"
    xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"
    xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
    <env:Header xmlns:env="http://www.w3.org/2003/05/soap-envelope">
        <wsa:Action>ConfigureResponse</wsa:Action>
        <wsa:MessageID>urn:uuid:9be492dc-a0b9-4cf9-908b-e02a11d623c3</wsa:MessageID>
        <wsa:RelatesTo>urn:uuid:b7df082c-4ddb-4a62-8cda-de290f42b9c0</wsa:RelatesTo>
        <wsa:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:To>
        <wsse:Security>
            <wsu:Timestamp wsu:Id="Timestamp-61ba3507-784d-4084-bdf0-0ca052fc29ec">
                <wsu:Created>2021-08-30T19:24:48Z</wsu:Created>
                <wsu:Expires>2021-08-30T19:29:48Z</wsu:Expires>
            </wsu:Timestamp>
        </wsse:Security>
    </env:Header>
    <soap:Body>
        <ConfigureResponseMsg xmlns="http://exacttarget.com/wsdl/partnerAPI">
            <Results>
                <Result>
                    <StatusCode>OK</StatusCode>
                    <StatusMessage>Success</StatusMessage>
                    <OrdinalID>0</OrdinalID>
                    <Object xsi:type="PropertyDefinition">
                        <PartnerKey xsi:nil="true" />
                        <ID>126713</ID>
                        <ObjectID xsi:nil="true" />
                        <Name>Test Subscriber Attribute</Name>
                        <PropertyType>string</PropertyType>
                    </Object>
                </Result>
            </Results>
            <OverallStatus>OK</OverallStatus>
            <OverallStatusMessage />
            <RequestID>ef959dde-a21d-46ac-b2de-98d9e656ed6f</RequestID>
        </ConfigureResponseMsg>
    </soap:Body>
</soap:Envelope>"""

CREATE_USER_EXPECTED_SOAP_RESPONSE = """<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd" xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
    <env:Header xmlns:env="http://www.w3.org/2003/05/soap-envelope">
        <wsa:Action>CreateResponse</wsa:Action>
        <wsa:MessageID>urn:uuid:73668493-d031-493c-b291-7fffe3bf46d0</wsa:MessageID>
        <wsa:RelatesTo>urn:uuid:9367afcb-c4ec-48f8-9a6f-2d0de4e37dff</wsa:RelatesTo>
        <wsa:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:To>
        <wsse:Security>
            <wsu:Timestamp wsu:Id="Timestamp-c58859fa-815f-4ecb-b278-d8df5249e0e6">
                <wsu:Created>2021-08-31T06:10:52Z</wsu:Created>
                <wsu:Expires>2021-08-31T06:15:52Z</wsu:Expires>
            </wsu:Timestamp>
        </wsse:Security>
    </env:Header>
    <soap:Body>
        <CreateResponse xmlns="http://exacttarget.com/wsdl/partnerAPI">
            <Results>
                <StatusCode>OK</StatusCode>
                <StatusMessage>Account User Updated / Created</StatusMessage>
                <OrdinalID>0</OrdinalID>
                <NewID>0</NewID>
                <Object xsi:type="AccountUser">
                    <Client>
                        <ID>523005197</ID>
                    </Client>
                    <PartnerKey xsi:nil="true" />
                    <ID>722602484</ID>
                    <ObjectID xsi:nil="true" />
                    <CustomerKey>Don_Draper_Key_1926</CustomerKey>
                    <UserID>sterling-don</UserID>
                    <Password />
                    <Name>Don Draper</Name>
                    <Email>don.draper@sterlingcooper.com</Email>
                    <UserPermissions>
                        <PartnerKey xsi:nil="true" />
                        <ID>31</ID>
                        <ObjectID xsi:nil="true" />
                        <Name xsi:nil="true" />
                        <Value xsi:nil="true" />
                        <Description xsi:nil="true" />
                        <Delete>0</Delete>
                    </UserPermissions>
                    <Delete>0</Delete>
                    <IsAPIUser>true</IsAPIUser>
                    <DefaultBusinessUnit>523008403</DefaultBusinessUnit>
                </Object>
            </Results>
            <RequestID>09647ed9-38d4-4431-9fa2-ae29df011a5a</RequestID>
            <OverallStatus>OK</OverallStatus>
        </CreateResponse>
    </soap:Body>
</soap:Envelope>"""

UPDATE_USER_ROLE_EXPECTED_SOAP_RESPONSE = """<?xml version="1.0" encoding="utf-8"?>
<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:wsa="http://schemas.xmlsoap.org/ws/2004/08/addressing" xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd" xmlns:wsu="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
    <env:Header xmlns:env="http://www.w3.org/2003/05/soap-envelope">
        <wsa:Action>CreateResponse</wsa:Action>
        <wsa:MessageID>urn:uuid:ed9644d9-3258-4799-899a-0fef9503d6b9</wsa:MessageID>
        <wsa:RelatesTo>urn:uuid:eaea2d53-ecf4-4ddc-b0c6-aa690d4141b1</wsa:RelatesTo>
        <wsa:To>http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous</wsa:To>
        <wsse:Security>
            <wsu:Timestamp wsu:Id="Timestamp-ee6031ae-b715-4ea0-a246-7607c807b830">
                <wsu:Created>2021-08-31T06:03:37Z</wsu:Created>
                <wsu:Expires>2021-08-31T06:08:37Z</wsu:Expires>
            </wsu:Timestamp>
        </wsse:Security>
    </env:Header>
    <soap:Body>
        <UpdateResponse xmlns="http://exacttarget.com/wsdl/partnerAPI">
            <Results>
                <StatusCode>OK</StatusCode>
                <StatusMessage>Account User Updated / Created</StatusMessage>
                <OrdinalID>0</OrdinalID>
                <Object xsi:type="AccountUser">
                    <Client>
                        <ID>523005197</ID>
                    </Client>
                    <PartnerKey xsi:nil="true" />
                    <ID>722602484</ID>
                    <ObjectID xsi:nil="true" />
                    <CustomerKey>Don_Draper_Key_1926</CustomerKey>
                    <AccountUserID>722602484</AccountUserID>
                    <Password />
                    <Name>Partner Don Draper</Name>
                    <Email>don.draper@sterlingcooper.com</Email>
                    <UserPermissions>
                        <PartnerKey xsi:nil="true" />
                        <ID>31</ID>
                        <ObjectID xsi:nil="true" />
                        <Name xsi:nil="true" />
                        <Value xsi:nil="true" />
                        <Description xsi:nil="true" />
                        <Delete>0</Delete>
                    </UserPermissions>
                    <Delete>0</Delete>
                    <IsAPIUser>true</IsAPIUser>
                </Object>
            </Results>
            <RequestID>37ee797a-12ae-4d99-8294-c9f3451c5cd5</RequestID>
            <OverallStatus>OK</OverallStatus>
        </UpdateResponse>
    </soap:Body>
</soap:Envelope>"""



import json
from unittest.mock import patch

import pytest
import responses

from cumulusci.core.exceptions import SalesforceException
from cumulusci.tasks.salesforce.composite import API_ROLLBACK_MESSAGE, CompositeApi

from .util import create_task

COMPOSITE_REQUEST = {
    "allOrNone": True,
    "compositeRequest": [
        {
            "method": "GET",
            "url": "/services/data/v46.0/query/?q=SELECT+Id+FROM+RecordType+WHERE+SobjectType+=+'Account' AND DeveloperName = 'Educational_Institution'",
            "referenceId": "schoolRt",
        },
        {
            "method": "POST",
            "url": "/services/data/v46.0/sobjects/Account",
            "referenceId": "uni",
            "body": {
                "Name": "Connected Campus University",
                "RecordTypeId": "@{schoolRt.records[0].Id}",
            },
        },
        {
            "method": "POST",
            "url": "/services/data/v46.0/sobjects/User",
            "referenceId": "sophiaUser",
            "body": {
                "FirstName": "Sophia",
                "LastName": "Student",
                "Alias": "sophia",
                "City": "San Francisco",
                "Country": "USA",
                "Email": "sofia@connected.edu",
                "EmailEncodingKey": "utf-8",
                "LanguageLocaleKey": "en_US",
                "LocaleSidKey": "en_US",
                "MobilePhone": "(650) 555-1212",
                "Phone": "(650) 555-1212",
                "PostalCode": "94105",
                "ProfileId": "@{profiles.records[0].Id}",
                "State": "CA",
                "Street": "1 Market Street",
                "TimeZoneSidKey": "America/Los_Angeles",
                "Username": "sofia@connected.edu",
            },
        },
    ],
}
COMPOSITE_RESPONSE = {
    "compositeResponse": [
        {
            "body": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "RecordType",
                            "url": "/services/data/v46.0/sobjects/RecordType/01211000002FEJ0AAO",
                        },
                        "Id": "01211000002FEJ0AAO",
                    }
                ],
            },
            "httpHeaders": {},
            "httpStatusCode": 200,
            "referenceId": "schoolRt",
        },
        {
            "body": {"id": "00111000021SzRwAAK", "success": True, "errors": []},
            "httpHeaders": {
                "Location": "/services/data/v46.0/sobjects/Account/00111000021SzRwAAK"
            },
            "httpStatusCode": 201,
            "referenceId": "uni",
        },
        {
            "body": {"id": "00511000009y7uaAAA", "success": True, "errors": []},
            "httpHeaders": {
                "Location": "/services/data/v46.0/sobjects/User/00511000009y7uaAAA"
            },
            "httpStatusCode": 201,
            "referenceId": "sophiaUser",
        },
    ]
}


class TestCompositeApi:
    @responses.activate
    @patch("cumulusci.tasks.salesforce.composite.CliTable")
    def test_composite_request(self, table, tmp_path):
        test_json = tmp_path / "test.json"
        test_json.write_text(json.dumps(COMPOSITE_REQUEST))
        task = create_task(
            CompositeApi,
            {
                "data_files": [
                    str(test_json),
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v52.0/composite",
            status=200,
            json=COMPOSITE_RESPONSE,
        )

        task()

        assert len(responses.calls) == 1
        request_body = json.loads(responses.calls[0].request.body)
        assert request_body == COMPOSITE_REQUEST

    @responses.activate
    @patch("cumulusci.tasks.salesforce.composite.CliTable")
    def test_composite_request_success_message(self, table, tmp_path):
        test_json = tmp_path / "test.json"
        test_json.write_text(json.dumps(COMPOSITE_REQUEST))
        task = create_task(
            CompositeApi,
            {
                "data_files": [
                    str(test_json),
                ],
            },
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v52.0/composite",
            status=200,
            json=COMPOSITE_RESPONSE,
        )
        expected_table_data = [
            ["ReferenceId", "Success"],
            ["schoolRt", True],
            ["uni", True],
            ["sophiaUser", True],
        ]

        task()
        table.assert_called_once_with(
            (expected_table_data), bool_cols=["Success"], title="Subrequest Results"
        )

    @responses.activate
    @patch("cumulusci.tasks.salesforce.composite.CliTable")
    def test_composite_request_exception(self, table, tmp_path):
        test_json = tmp_path / "test.json"
        test_json.write_text(json.dumps(COMPOSITE_REQUEST))
        task = create_task(
            CompositeApi,
            {
                "data_files": [
                    str(test_json),
                ],
            },
        )

        error_response = {
            "compositeResponse": [
                {
                    "body": [
                        {
                            "errorCode": "PROCESSING_HALTED",
                            "message": "Invalid reference specified. No value for schoolRt.records[0].Id found in schoolRt",
                        }
                    ],
                    "httpHeaders": {},
                    "httpStatusCode": 400,
                    "referenceId": "badref",
                },
                {
                    "body": [
                        {
                            "errorCode": "PROCESSING_HALTED",
                            "message": API_ROLLBACK_MESSAGE,
                        }
                    ],
                    "httpHeaders": {},
                    "httpStatusCode": 400,
                    "referenceId": "rollback",
                },
            ]
        }
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v52.0/composite",
            status=200,
            json=error_response,
        )
        expected_table_data = [
            ["ReferenceId", "Message"],
            [
                "badref",
                "Invalid reference specified. No value for schoolRt.records[0].Id found in schoolRt",
            ],
        ]

        with pytest.raises(SalesforceException):
            task()

        table.assert_called_once_with((expected_table_data), wrap_cols=["Message"])

    def test_json_processing(self):
        request = COMPOSITE_REQUEST["compositeRequest"].copy()
        request.append(
            {
                "method": "PATCH",
                "url": "/services/data/v46.0/sobjects/User",
                "body": {
                    "Id": "%%%USERID%%%",
                    "Email": "test@testerino.patch",
                    "Test__c": "%%%NAMESPACE%%%Foo__c",
                },
            }
        )
        body = json.dumps(request)
        task = create_task(
            CompositeApi,
            {
                "data_files": [
                    "dummy_path",
                ],
                "randomize_username": True,
                "managed": True,
            },
        )

        task.project_config.project__package__namespace = "NS"
        processed_body = task._process_json(body)
        assert "USER_ID" in processed_body
        assert "NS__Foo__c" in processed_body

        assert '"Username": "sofia@connected.' in processed_body
        assert '"Username": "sofia@connected.edu"' not in processed_body



from cumulusci.core.keychain.encrypted_file_project_keychain import (
    EncryptedFileProjectKeychain,
)

# All logic for loading orgs and services from the environment
# now lives in EncryptedFileProjectKeychain.
EnvironmentProjectKeychain = EncryptedFileProjectKeychain  # pragma: no cover



from collections import defaultdict
from typing import Dict

import click
import yaml

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.salesforce_api.org_schema import Field, get_org_schema
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class GenerateMapping(BaseSalesforceApiTask):
    task_docs = """
    Generate a mapping file for use with the `extract_dataset` and `load_dataset` tasks.
    This task will examine the schema in the specified org and attempt to infer a
    mapping suitable for extracting data in packaged and custom objects as well as
    customized standard objects.

    Mappings must be serializable, and hence must resolve reference cycles - situations
    where Object A refers to B, and B also refers to A. Mapping generation will stop
    and request user input to resolve such cycles by identifying the correct load order.
    If you would rather the mapping generator break such a cycle randomly, set the
    `break_cycles` option to `auto`.

    Alternately, specify the `ignore` option with the name of one of the
    lookup fields to suppress it and break the cycle. `ignore` can be specified as a list in
    `cumulusci.yml` or as a comma-separated string at the command line.

    In most cases, the mapping generated will need minor tweaking by the user. Note
    that the mapping omits features that are not currently well supported by the
    `extract_dataset` and `load_dataset` tasks, such as references to
    the `User` object.
    """

    task_options = {
        "path": {"description": "Location to write the mapping file", "required": True},
        "namespace_prefix": {"description": "The namespace prefix to use"},
        "ignore": {
            "description": "Object API names, or fields in Object.Field format, to ignore"
        },
        "break_cycles": {
            "description": "If the generator is unsure of the order to load, what to do? "
            "Set to `ask` (the default) to allow the user to choose or `auto` to pick randomly."
        },
        "include": {
            "description": "Object names to include even if they might not otherwise be included."
        },
        "strip_namespace": {
            "description": "If True, CumulusCI removes the project's namespace where found in fields "
            " and objects to support automatic namespace injection. On by default."
        },
    }

    core_fields = ["Name", "FirstName", "LastName"]

    def _init_options(self, kwargs):
        super(GenerateMapping, self)._init_options(kwargs)
        if "namespace_prefix" not in self.options:
            self.options["namespace_prefix"] = ""

        if self.options["namespace_prefix"] and not self.options[
            "namespace_prefix"
        ].endswith("__"):
            self.options["namespace_prefix"] += "__"

        self.options["ignore"] = process_list_arg(self.options.get("ignore", []))
        break_cycles = self.options.setdefault("break_cycles", "ask")
        if break_cycles not in ["ask", "auto"]:
            raise TaskOptionsError(
                f"`break_cycles` should be `ask` or `auto`, not {break_cycles}"
            )
        self.options["include"] = process_list_arg(self.options.get("include", []))
        strip_namespace = self.options.get("strip_namespace")

        self.options["strip_namespace"] = process_bool_arg(
            True if strip_namespace is None else strip_namespace
        )

    def _run_task(self):
        self.logger.info("Collecting sObject information")
        with get_org_schema(self.sf, self.org_config) as org_schema:
            self._collect_objects(org_schema)
            self._simplify_schema(org_schema)
        filename = self.options["path"]
        self.logger.info(f"Creating mapping schema {filename}")
        self._build_mapping()
        with open(filename, "w") as f:
            yaml.dump(self.mapping, f, sort_keys=False)

    def _collect_objects(self, org_schema):
        """Walk the global describe and identify the sObjects we need to include in a minimal operation."""
        self.mapping_objects = self.options["include"]

        sobject_names = set(org_schema.keys())

        unknown_objects = set(self.mapping_objects) - sobject_names

        if unknown_objects:
            raise TaskOptionsError(f"{unknown_objects} cannot be found in the org.")

        # First, we'll get a list of all objects that are either
        # (a) custom, no namespace
        # (b) custom, with our namespace
        # (c) not ours (standard or other package), but have fields with our namespace or no namespace
        for objname, obj in org_schema.items():
            if self._is_our_custom_api_name(objname) or self._has_our_custom_fields(
                obj
            ):
                if (
                    self._is_object_mappable(obj)
                    and objname not in self.mapping_objects
                ):
                    self.mapping_objects.append(objname)

        # Add any objects that are required by our own,
        # meaning any object we are looking up to with a custom field,
        # or any master-detail parent of any included object.
        index = 0
        while index < len(self.mapping_objects):
            obj = self.mapping_objects[index]
            for field in org_schema[obj].fields.values():
                if field["type"] == "reference":
                    if field["relationshipOrder"] == 1 or self._is_any_custom_api_name(
                        field["name"]
                    ):
                        self.mapping_objects.extend(
                            [
                                obj
                                for obj in field["referenceTo"]
                                if obj not in self.mapping_objects
                                and self._is_object_mappable(org_schema[obj])
                            ]
                        )

            index += 1

    def _simplify_schema(self, org_schema):
        """Simplify and filter schema, including field details and interobject
        references, into self.simple_schema and self.refs"""

        # Now, find all the fields we need to include.
        # For custom objects, we include all custom fields. This includes custom objects
        # that our package doesn't own.
        # For standard objects, we include all custom fields, all required standard fields,
        # and master-detail relationships. Required means createable and not nillable.
        # In all cases, ensure that RecordTypeId is included if and only if there are Record Types
        self.simple_schema = {}
        self.refs = defaultdict(lambda: defaultdict(dict))
        for obj in self.mapping_objects:
            self.simple_schema[obj] = {}

            for field in org_schema[obj]["fields"].values():
                if any(
                    [
                        self._is_any_custom_api_name(field["name"]),
                        self._is_core_field(field["name"]),
                        self._is_required_field(field),
                        self._is_lookup_to_included_object(field),
                    ]
                ):
                    if self._is_field_mappable(obj, field):
                        self.simple_schema[obj][field["name"]] = field

                        if field["type"] == "reference":
                            for target in field["referenceTo"]:
                                # We've already vetted that this field is referencing
                                # included objects, via `_is_field_mappable()`
                                if target != obj:
                                    self.refs[obj][target][field["name"]] = field

                if (
                    field["name"] == "RecordTypeId"
                    and org_schema[obj].recordTypeInfos
                    and len(org_schema[obj].recordTypeInfos) > 1
                ):
                    # "Master" is included even if no RTs.c
                    self.simple_schema[obj][field["name"]] = field

    def _build_mapping(self):
        """Output self.simple_schema in mapping file format by constructing a dict and serializing to YAML"""
        objs = list(self.simple_schema.keys())

        stack = self._split_dependencies(objs, self.refs)
        ns = self.project_config.project__package__namespace

        def strip_namespace(element):
            if self.options["strip_namespace"] and ns and element.startswith(f"{ns}__"):
                return element[len(ns) + 2 :]
            else:
                return element

        self.mapping = {}
        for orig_obj in stack:
            # Check if it's safe for us to strip the namespace from this object
            stripped_obj = strip_namespace(orig_obj)
            obj = stripped_obj if stripped_obj not in stack else orig_obj
            key = f"Insert {obj}"
            self.mapping[key] = {}
            self.mapping[key]["sf_object"] = obj
            fields = []
            lookups = []
            for field in self.simple_schema[orig_obj].values():
                if field["type"] == "reference" and field["name"] != "RecordTypeId":
                    # For lookups, namespace stripping takes place below.
                    lookups.append(field["name"])
                else:
                    fields.append(field["name"])
            if fields:
                fields_stripped = [
                    strip_namespace(f) if strip_namespace(f) not in fields else f
                    for f in fields
                ]
                fields_stripped.sort()
                self.mapping[key]["fields"] = fields_stripped
            if lookups:
                lookups.sort()
                self.mapping[key]["lookups"] = {}
                for orig_field in lookups:
                    # First, determine what manner of lookup we have here.
                    stripped_field = (
                        strip_namespace(orig_field)
                        if strip_namespace(orig_field) not in lookups
                        else orig_field
                    )
                    referenceTo = self.simple_schema[orig_obj][orig_field][
                        "referenceTo"
                    ]

                    if len(referenceTo) > 1:  # Polymorphic lookup
                        self.logger.warning(
                            f"Field {orig_obj}.{orig_field} is a polymorphic lookup, which is not supported"
                        )
                    else:
                        orig_reference = referenceTo[0]

                        # Can we safely namespace-strip this reference?
                        stripped_reference = (
                            strip_namespace(orig_reference)
                            if strip_namespace(orig_reference) not in stack
                            else orig_reference
                        )

                        if orig_reference == orig_obj:  # Self-lookup
                            self.mapping[key]["lookups"][stripped_field] = {
                                "table": stripped_reference,
                                "after": key,
                            }
                        elif stack.index(orig_reference) > stack.index(
                            orig_obj
                        ):  # Dependent lookup
                            self.mapping[key]["lookups"][stripped_field] = {
                                "table": stripped_reference,
                                "after": f"Insert {stripped_reference}",
                            }
                        else:  # Regular lookup
                            self.mapping[key]["lookups"][stripped_field] = {
                                "table": stripped_reference
                            }

    def _split_dependencies(self, objs, dependencies):
        """Attempt to flatten the object network into a sequence of load operations."""
        stack = []
        objs_remaining = sorted(objs)

        # The structure of `dependencies` is:
        # key = object, value = set of objects it references.

        # Iterate through our list of objects
        # For each object, if it is not dependent on any other objects, place it at the end of the stack.
        # Once an object is placed in the stack, remove dependencies to it (they're satisfied)
        while objs_remaining:
            objs_without_deps = [
                obj
                for obj in objs_remaining
                if obj not in dependencies or not dependencies[obj]
            ]
            assert all(objs_without_deps)

            if not objs_without_deps:
                choice = self.choose_next_object(objs_remaining, dependencies)
                assert choice
                objs_without_deps = [choice]

            for obj in objs_without_deps:
                stack.append(obj)

                # Remove all dependencies on this object (they're satisfied)
                for other_obj in dependencies:
                    if obj in dependencies.get(other_obj):
                        del dependencies[other_obj][obj]

                # Remove this object from our remaining set.
                objs_remaining.remove(obj)

        return stack

    def find_free_object(self, objs_remaining: list, dependencies: dict):
        # if you change this code, remember that
        # peeking into a generator consumes it
        free_objs = (
            sobj
            for sobj in objs_remaining
            if only_has_soft_dependencies(sobj, dependencies[sobj])
        )
        first_free_obj = next(free_objs, None)

        return first_free_obj

    def choose_next_object(self, objs_remaining: list, dependencies: dict):
        free_obj = self.find_free_object(objs_remaining, dependencies)
        if free_obj:
            return free_obj

        if self.options["break_cycles"] == "auto":
            return tuple(objs_remaining)[0]
        else:
            return self.ask_user(objs_remaining, dependencies)

    def ask_user(self, objs_remaining, dependencies):
        self.logger.info(
            "CumulusCI needs help to complete the mapping; the schema contains reference cycles and unresolved dependencies."
        )
        self.logger.info("Remaining objects:")
        for obj in objs_remaining:
            self.logger.info(obj)
            for other_obj in dependencies[obj]:
                self.logger.info(
                    f"   references {other_obj} via: {', '.join(dependencies[obj][other_obj])}"
                )

        return click.prompt(
            "Which object should we load first?",
            type=click.Choice(tuple(objs_remaining)),
            show_choices=True,
        )

    def _is_any_custom_api_name(self, api_name: str):
        """True if the entity name is custom (including any package)."""
        return api_name.endswith("__c")

    def _is_our_custom_api_name(self, api_name: str):
        """True if the entity name is custom and has our namespace prefix (if we have one)
        or if the entity does not have a namespace"""
        return self._is_any_custom_api_name(api_name) and (
            (
                self.options["namespace_prefix"]
                and api_name.startswith(self.options["namespace_prefix"])
            )
            or api_name.count("__") == 1
        )

    def _is_core_field(self, api_name):
        """True if this field is one that we should always include regardless
        of other settings or field configuration, such as Contact.FirstName.
        DB-level required fields don't need to be so handled."""

        return api_name in self.core_fields

    def _is_object_mappable(self, obj):
        """True if this object is one we can map, meaning it's an sObject and not
        some other kind of entity, it's not ignored, it's Bulk API compatible,
        and it's not in a hard-coded list of entities we can't currently handle."""

        return not any(
            [
                obj["name"] in self.options["ignore"],  # User-specified exclusions
                obj["name"].endswith(
                    "ChangeEvent"
                ),  # Change Data Capture entities (which get custom fields)
                obj["name"].endswith("__mdt"),  # Custom Metadata Types (MDAPI only)
                obj["name"].endswith("__e"),  # Platform Events
                obj["customSetting"],  # Not Bulk API compatible
                obj["name"]  # Objects we can't or shouldn't load/save
                in [
                    "User",
                    "Group",
                    "LookedUpFromActivity",
                    "OpenActivity",
                    "Task",
                    "Event",
                    "ActivityHistory",
                ],
            ]
        )

    def _is_field_mappable(self, obj, field):
        """True if this field is one we can map, meaning it's not ignored,
        it's createable by the Bulk API, it's not a deprecated field,
        and it's not a type of reference we can't handle without special
        configuration (self-lookup or reference to objects not included
        in this operation)."""
        return not any(
            [
                field["name"] == "Id",  # Omit Id fields for auto-pks
                f"{obj}.{field['name']}" in self.options["ignore"],  # User-ignored list
                "(Deprecated)" in field["label"],  # Deprecated managed fields
                field["type"] == "base64",  # No Bulk API support for base64 blob fields
                not field["createable"],  # Non-writeable fields
                field["type"] == "reference"  # Outside lookups
                and not self._are_lookup_targets_in_operation(field),
            ]
        )

    def _is_required_field(self, field):
        """True if the field is either database-level required or a master-detail
        relationship field."""
        return (field["createable"] and not field["nillable"]) or (
            field["type"] == "reference" and field["relationshipOrder"] == 1
        )

    def _has_our_custom_fields(self, obj):
        """True if the object is owned by us or contains any field owned by us."""
        return any(
            [self._is_our_custom_api_name(fieldname) for fieldname in obj.fields]
        )

    def _are_lookup_targets_in_operation(self, field):
        """True if this lookup field aims at objects we are already including (all targets
        must match, although we don't provide actual support for polymorphism)."""
        return all([f in self.mapping_objects for f in field["referenceTo"]])

    def _is_lookup_to_included_object(self, field):
        """True if this field is a lookup and also references only objects we are
        already including."""
        return field["type"] == "reference" and self._are_lookup_targets_in_operation(
            field
        )


def only_has_soft_dependencies(
    sobj: str, obj_dependencies: Dict[str, Dict[str, Field]]
):
    for target_obj, field_deps in obj_dependencies.items():
        for field_name, field_data in field_deps.items():
            # all nillable references are considered soft dependencies.
            #
            # A single hard dependency renders an object "not yet free"
            if not field_data.nillable:
                return False

    return True



from cumulusci.salesforce_api.package_zip import CreatePackageZipBuilder
from cumulusci.tasks.salesforce import Deploy


class CreatePackage(Deploy):
    task_options = {
        "package": {
            "description": "The name of the package to create.  Defaults to project__package__name",
            "required": True,
        },
        "api_version": {
            "description": "The api version to use when creating the package.  Defaults to project__package__api_version",
            "required": True,
        },
    }

    def _init_options(self, kwargs):
        super(CreatePackage, self)._init_options(kwargs)
        if "package" not in self.options:
            self.options["package"] = self.project_config.project__package__name
        if "api_version" not in self.options:
            self.options[
                "api_version"
            ] = self.project_config.project__package__api_version

    def _get_package_zip(self, path=None):
        return CreatePackageZipBuilder(
            self.options["package"], self.options["api_version"]
        ).as_base64()



""" Tests for the SFDX Command Wrapper"""

import unittest
from unittest import mock
from unittest.mock import MagicMock, patch

from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    ScratchOrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.command import CommandException
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.sfdx import SFDXBaseTask, SFDXJsonTask, SFDXOrgTask


class TestSFDXBaseTask(MockLoggerMixin, unittest.TestCase):
    """Tests for the Base Task type"""

    def setUp(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig()

        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_base_task(self):
        """The command is prefixed w/ sfdx"""

        self.task_config.config["options"] = {"command": "force:org", "extra": "--help"}
        task = SFDXBaseTask(self.project_config, self.task_config)

        try:
            task()
        except CommandException:
            pass

        self.assertEqual("force:org", task.options["command"])
        self.assertEqual("sfdx force:org --help", task._get_command())

    @patch("cumulusci.tasks.command.Command._run_task", MagicMock(return_value=None))
    def test_keychain_org_creds(self):
        """Keychain org creds are passed by env var"""

        self.task_config.config["options"] = {"command": "force:org --help"}
        access_token = "00d123"
        org_config = OrgConfig(
            {
                "instance_url": "https://test.salesforce.com",
            },
            "test",
        )

        def refresh_oauth_token(keychain):
            org_config.config["access_token"] = access_token

        org_config.refresh_oauth_token = mock.Mock(side_effect=refresh_oauth_token)
        org_config.save = mock.Mock()

        task = SFDXOrgTask(self.project_config, self.task_config, org_config)
        task()

        org_config.refresh_oauth_token.assert_called_once()
        self.assertIn("SFDX_INSTANCE_URL", task._get_env())
        self.assertIn("SFDX_DEFAULTUSERNAME", task._get_env())
        self.assertIn(access_token, task._get_env()["SFDX_DEFAULTUSERNAME"])

    def test_scratch_org_username(self):
        """Scratch Org credentials are passed by -u flag"""
        self.task_config.config["options"] = {"command": "force:org --help"}
        org_config = ScratchOrgConfig({"username": "test@example.com"}, "test")

        task = SFDXOrgTask(self.project_config, self.task_config, org_config)
        self.assertIn("-u test@example.com", task._get_command())


class TestSFDXJsonTask(unittest.TestCase):
    def test_get_command(self):
        task = create_task(SFDXJsonTask)
        command = task._get_command()
        self.assertEqual("sfdx force:mdapi:deploy --json", command)

    def test_process_output(self):
        task = create_task(SFDXJsonTask)
        task.logger = mock.Mock()
        task._process_output("{}")
        task.logger.info.assert_called_once_with("JSON = {}")

    def test_process_output__json_parse_error(self):
        task = create_task(SFDXJsonTask)
        task.logger = mock.Mock()
        task._process_output("{")
        task.logger.error.assert_called_once()



import logging
import time

logger = logging.getLogger(__name__)


def retry(
    func,
    should_retry=lambda e: True,
    retries: int = 5,
    retry_interval: int = 5,
    retry_interval_add: int = 30,
):
    while True:
        try:
            return func()
            break
        except Exception as e:
            if not (retries and should_retry(e)):
                raise
            if retry_interval:
                logger.warning(f"Sleeping for {retry_interval} seconds before retry...")
                time.sleep(retry_interval)
                if retry_interval_add:
                    retry_interval += retry_interval_add
            retries -= 1
            logger.warning(f"Retrying ({retries} attempts remaining)")


def poll(action):
    """poll for a result in a loop"""
    count = 0
    interval = 1
    while True:
        count += 1
        complete = action()
        if complete:
            break
        time.sleep(interval)
        if count % 3 == 0:
            interval += 1
